#include <windows.h>
#include <commctrl.h>
#include <strsafe.h>
#include "resource.h"

/* whichFormat determines the line break format of a file through statistical analysis. If
 * over 90% of the line breaks in a file are in a particular format, the function
 * indicates that it is the format of the file. Otherwise, the function reports failure.
 * Returns 1 if Windows format, 2 if Unix format, 3 if Mac format, or 0 on failure
 */
int whichFormat(char * filePath);

/* replaceStr replaces all occurrences of oldS in in with newS and stores it in out. in,
 * oldS, and newS are unchanged as are the values they point to. out and the value it
 * points to are both changed. out must be a reference to a pointer rather than just a
 * pointer because replaceStr allocates memory and the new value of out that points to the
 * memory must be reflected in the calling function.
 * Returns nonzero on success, zero on failure
 * ANSI Only
 */
int replaceStr(const char * in, const char * oldS, const char * newS, char * & out);

/* About Box Dialog Procedure
 * This function processes Close and Ok buttons and closes window when these buttons are
 * pushed.
 */
INT_PTR CALLBACK AboutDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
switch(uMsg) {
	case WM_INITDIALOG:
	break;
	case WM_COMMAND:
	switch(LOWORD(wParam)) {
		case IDOK:
		EndDialog(hwndDlg, IDOK);
		break;
	}
	break;
	case WM_CLOSE:
	EndDialog(hwndDlg, 0);
	break;
	default:
	return false;
}
return true;
}

/* Main Dialog Procedure
 * This basically handles anything of any importance for the application. It will pass
 * control to the replaceStr function when necessary in order to comply with conversion
 * requests.
 */
INT_PTR CALLBACK MainDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
switch(uMsg) {
	case WM_INITDIALOG:
	// Set the main window icon to the program icon
	SendMessage(hwndDlg, WM_SETICON, ICON_SMALL, (LPARAM)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_PROGICON), IMAGE_ICON, 16, 16, 0));
	SendMessage(hwndDlg, WM_SETICON, ICON_BIG, (LPARAM)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_PROGICON), IMAGE_ICON, 32, 32, 0));
	
	/* Make sure that someone doesn't try to type in a file with a path longer than
	 * MAX_PATH, although that would be quite a long path to type in since EM_LIMITTEXT
	 * only affects typed-in text.
	 */
	SendDlgItemMessage(hwndDlg, IDE_INFILE, EM_LIMITTEXT, (WPARAM)MAX_PATH, 0);
	SendDlgItemMessage(hwndDlg, IDE_OUTFILE, EM_LIMITTEXT, (WPARAM)MAX_PATH, 0);
	break;
	case WM_COMMAND:
	switch(LOWORD(wParam)) {
		case IDB_IN_SEL: // User wants to choose input file
		{
			char inPath[MAX_PATH] = ""; // Need to initialize array for OPENFILENAME structure.
			
			OPENFILENAME ofn; // Needed for GetOpenFileName
			ZeroMemory(&ofn, sizeof(OPENFILENAME)); // Clear memory.
			ofn.lStructSize = sizeof(OPENFILENAME); // Make required record of size.
			
			// Set starting paramaters for the Open dialog box.
			ofn.hwndOwner = hwndDlg;
			ofn.lpstrFilter = "Text Files (*.txt)\0*.txt\0All Files (*.*)\0*.*\0";
			ofn.nMaxFile = MAX_PATH;
			ofn.lpstrDefExt = "txt";
			ofn.lpstrFile = inPath;
			ofn.Flags = OFN_EXPLORER | OFN_HIDEREADONLY; // Use new style dialog and don't show "Open Read-Only" checkbox.
			
			// All set, now make open dialog box. If successful, record file path and continue. Otherwise, stop.
			if(GetOpenFileName(&ofn)) SendDlgItemMessage(hwndDlg, IDE_INFILE, WM_SETTEXT, 0, (LPARAM)inPath);
			else break;
			
			// Now check to see the format of the input file and set dialog controls accordingly.
			switch(whichFormat(inPath)) {
				case 1: // It's in windows format. Set Windows input radio button.
				EnableWindow(GetDlgItem(hwndDlg, IDB_IN_WIN), true);
				SendDlgItemMessage(hwndDlg, IDB_OUT_WIN, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
				SendDlgItemMessage(hwndDlg, IDB_IN_WIN, BM_CLICK, 0, 0);
				break;
				case 2: // It's in Unix format. Set Unix input radio button.
				EnableWindow(GetDlgItem(hwndDlg, IDB_IN_UNIX), true);
				SendDlgItemMessage(hwndDlg, IDB_OUT_UNIX, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
				SendDlgItemMessage(hwndDlg, IDB_IN_UNIX, BM_CLICK, 0, 0);
				break;
				case 3: // It's in Mac format. Set Mac input radio button.
				EnableWindow(GetDlgItem(hwndDlg, IDB_IN_MAC), true);
				SendDlgItemMessage(hwndDlg, IDB_OUT_MAC, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
				SendDlgItemMessage(hwndDlg, IDB_IN_MAC, BM_CLICK, 0, 0);
				break;
				default: // Couldn't determine the format. Inform the user and don't touch anything.
				MessageBox(hwndDlg, "Couldn't automatically detect text format.\nPlease select one manually.", "Alert", MB_ICONEXCLAMATION | MB_OK);
				break;
			}
		}
		break;
		case IDB_OUT_SEL: // User wants to choose the output file.
		{
			char outPath[MAX_PATH] = ""; // Need to initialize array for OPENFILENAME structure.
			
			OPENFILENAME ofn; // Needed for GetSaveFileName
			ZeroMemory(&ofn, sizeof(OPENFILENAME)); // Clear memory.
			ofn.lStructSize = sizeof(OPENFILENAME); // Make required record of size.
			
			// Set starting paramaters for the Open dialog box.
			ofn.hwndOwner = hwndDlg;
			ofn.lpstrFilter = "Text Files (*.txt)\0*.txt\0All Files (*.*)\0*.*\0";
			ofn.nMaxFile = MAX_PATH;
			ofn.lpstrDefExt = "txt";
			ofn.lpstrFile = outPath;
			ofn.Flags = OFN_EXPLORER | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY; // Use new style dialog and don't show "Open Read-Only" checkbox.
			
			// All set, now make save dialog box. If successful, record file path.
			if(GetSaveFileName(&ofn)) SendDlgItemMessage(hwndDlg, IDE_OUTFILE, WM_SETTEXT, 0, (LPARAM)outPath);
		}
		break;
		/* The following six cases are designed to fire when the user clicks a radio
		 * button. When this happens, the opposite radio button (e.g. the Windows output
		 * radio button, if the Windows input radio button was clicked) will be disabled,
		 * and the sibling radio buttons (e.g. the other input buttons, using the previous
		 * example) will be enabled (if they were disabled previously).
		 */
		case IDB_IN_WIN:
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_WIN), false);
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_UNIX), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_MAC), true);
		break;
		case IDB_IN_UNIX:
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_WIN), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_UNIX), false);
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_MAC), true);
		break;
		case IDB_IN_MAC:
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_WIN), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_UNIX), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_OUT_MAC), false);
		break;
		case IDB_OUT_WIN:
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_WIN), false);
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_UNIX), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_MAC), true);
		break;
		case IDB_OUT_UNIX:
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_WIN), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_UNIX), false);
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_MAC), true);
		break;
		case IDB_OUT_MAC:
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_WIN), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_UNIX), true);
		EnableWindow(GetDlgItem(hwndDlg, IDB_IN_MAC), false);
		break;
		case IDB_CONVERT: // The user is ready to convert the file now.
		{
			/* The following variables store the address of the string that defines the
			 * line break format of the old and new files. Note that the pointers are not
			 * constant, but the memory they point to is.
			 */
			const char * oldF; 
			const char * newF;
			
			// We need these buffers to store the file paths.
			char inPath[MAX_PATH] = "";
			char outPath[MAX_PATH] = "";
			
			// Fetch the paths from the edit boxes.
			SendDlgItemMessage(hwndDlg, IDE_INFILE, WM_GETTEXT, (WPARAM)MAX_PATH, (LPARAM)inPath);
			SendDlgItemMessage(hwndDlg, IDE_OUTFILE, WM_GETTEXT, (WPARAM)MAX_PATH, (LPARAM)outPath);
			
			// Fetch the input proper format from the radio boxes and set the oldF pointer accordingly.
			if (SendDlgItemMessage(hwndDlg, IDB_IN_WIN, BM_GETCHECK, 0, 0) == BST_CHECKED) {
				oldF = "\r\n";
			} else if (SendDlgItemMessage(hwndDlg, IDB_IN_UNIX, BM_GETCHECK, 0, 0) == BST_CHECKED) {
				oldF = "\n";
			} else if (SendDlgItemMessage(hwndDlg, IDB_IN_MAC, BM_GETCHECK, 0, 0) == BST_CHECKED) {
				oldF = "\r";
			} else {
				// No input format, stop conversion.
				MessageBox(hwndDlg, "No input format was specified!", "Error", MB_OK | MB_ICONERROR);
				break;
			}
			
			// Fetch the output proper format from the radio boxes and set the newF pointer accordingly.
			if (SendDlgItemMessage(hwndDlg, IDB_OUT_WIN, BM_GETCHECK, 0, 0) == BST_CHECKED) {
				newF = "\r\n";
			} else if (SendDlgItemMessage(hwndDlg, IDB_OUT_UNIX, BM_GETCHECK, 0, 0) == BST_CHECKED) {
				newF = "\n";
			} else if (SendDlgItemMessage(hwndDlg, IDB_OUT_MAC, BM_GETCHECK, 0, 0) == BST_CHECKED) {
				newF = "\r";
			} else {
				// No output format, stop conversion.
				MessageBox(hwndDlg, "No output format was specified!", "Error", MB_OK | MB_ICONERROR);
				break;
			}
			
			// All set, open input file handle.
			HANDLE inFile = CreateFile(inPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_ARCHIVE, NULL);
			
			// Error, stop.
			if (inFile == INVALID_HANDLE_VALUE) {
				MessageBox(hwndDlg, "Could not open the input file!", "Error", MB_OK | MB_ICONERROR);
				break;
			}
			
			// Get file size to determine size of buffer.
			unsigned long inLenExt;
			unsigned long inLen = GetFileSize(inFile, &inLenExt);
			
			// File is too large for us to handle. Very unlikely for a plain text file.
			if (inLenExt != 0) {
				MessageBox(NULL, "File Conversion Error:\nInput file too large!", "Error", MB_OK | MB_ICONERROR);
			} else {
				char * inText = new char[inLen + 1]; // Allocate buffer for input file.
				
				// Read file into buffer
				DWORD dwRead;
				ReadFile(inFile, inText, inLen, &dwRead, NULL);
				inText[inLen] = '\0';
				
				// Make pointer for output text.
				char * outText = 0;
				
				// Convert text and check for errors.
				if (!replaceStr(inText, oldF, newF, outText)) {
					MessageBox(NULL, "File Conversion Error", "Error", MB_OK | MB_ICONERROR);
				} else {
					// Now get length of output.
					size_t outLen;
					StringCbLength(outText, STRSAFE_MAX_CCH, &outLen);
					
					// Create output file.
					HANDLE outFile  = CreateFile(outPath, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL);
					
					// Check for errors.
					if (outFile == INVALID_HANDLE_VALUE) {
						MessageBox(hwndDlg, "Could not create the output file!", "Error", MB_OK | MB_ICONERROR);
					} else {
						// No errors, so write the output file.
						WriteFile(outFile, outText, outLen, &dwRead, NULL);
						
						// We're finally done.
						MessageBox(NULL, "File converted successfully!", "Success", MB_OK | MB_ICONINFORMATION);
					}
					delete [] outText; // Free output buffer
					
					CloseHandle(outFile); // Free output handle
				}
				delete [] inText; // Free input buffer
			}
			CloseHandle(inFile); // Free input handle
		}
		break;
		case IDB_ABOUT: // User wants about dialog, so show it and check for errors.
		if (DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT), hwndDlg, AboutDlgProc) == -1) {
			MessageBox(hwndDlg, "About dialog failed!", "Error", MB_OK | MB_ICONINFORMATION);
		}
		break;
		case IDB_EXIT: // User wants to exit.
		EndDialog(hwndDlg, 0); // End dialog, also exits program.
		break;
	}
	break;
	case WM_CLOSE: // User wants to exit.
	EndDialog(hwndDlg, 0); // End dialog, also exits program.
	break;
	default:
	return false;
}
return true;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
// Begin initializing common controls.
INITCOMMONCONTROLSEX icc;

icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
icc.dwICC = 0;

InitCommonControlsEx(&icc);
// Done initializing common controls.

if (DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, MainDlgProc) == -1) { // Make main dialog.
MessageBox(NULL, "Creation of main window failed!", "Error", MB_OK | MB_ICONERROR);
}

return 0; // Good day, sir.
}

// See notes on whichFormat at prototype.
int whichFormat(char * filePath) {
int i; // for index, to be used later

// Open the file passed as an argument.
HANDLE file = CreateFile(filePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_ARCHIVE, NULL);

// If file couldn't be opened, exit function, return error.
if (file == INVALID_HANDLE_VALUE) return 0;

// Get file size to determine size of buffer.
unsigned long inLenExt;
unsigned long inLen = GetFileSize(file, &inLenExt);

// File is too large for us to handle - return error. Very unlikely for a plain text file.
if(inLenExt > 0) {
	CloseHandle(file);
	return 0;
}

// Allocate buffer for file.
char * inText = new char[inLen + 1];

// Read file into buffer.
DWORD dwRead;
ReadFile(file, inText, inLen, &dwRead, NULL);
inText[inLen] = '\0';

// Done with the file, so close it.
CloseHandle(file);

// Start counting the number of each type of line break.
int winCount = 0, unixCount = 0, macCount = 0;
for (i=0;i<inLen;i++) {
	if (inText[i] == '\r' && inText[i + 1] == '\n') winCount++;
	if (inText[i] != '\r' && inText[i + 1] == '\n') unixCount++; // Check if there is a unix line break AND NOT a Windows line break here.
	if (inText[i] == '\r' && inText[i + 1] != '\n') macCount++; // Check if there is a mac line break AND NOT a Windows line break here.
}

// We're done with the input text, so free the memory.
delete [] inText;

// Calculate the total number of line breaks in the file.
int totalCount = winCount + unixCount + macCount;

// If any one type of line break occurs more than 90% (.9) of the time, consider it to be the format that represents the file type.
if ((float)winCount / (float)totalCount > 0.9) return 1; // It's a windows file.
if ((float)unixCount / (float)totalCount > 0.9) return 2; // It's a Unix file.
if ((float)macCount / (float)totalCount > 0.9) return 3; // It's a Mac file.

return 0;
}

// See notes on replaceStr at prototype.
int replaceStr(const char * in, const char * oldS, const char * newS, char * & out) {
unsigned int i, j; // Looping indexes
size_t inLen, oldSLen, newSLen; // Argument lengths

// Check input argument.
if (FAILED(StringCbLength(in, STRSAFE_MAX_CCH, &inLen))) {
	MessageBox(NULL, "Failed to read input properly!\nThis could be because it is simply too large.", "Error", MB_ICONERROR | MB_OK);
}
// Check oldS argument.
if (FAILED(StringCbLength(oldS, STRSAFE_MAX_CCH, &oldSLen))) {
	MessageBox(NULL, "Program Error!\nPlease notify the developer: replaceStr oldS error.", "Error", MB_ICONERROR | MB_OK);
}
// Check newS argument.
if (FAILED(StringCbLength(newS, STRSAFE_MAX_CCH, &newSLen))) {
	MessageBox(NULL, "Program Error!\nPlease notify the developer: replaceStr newS error.", "Error", MB_ICONERROR | MB_OK);
}
if (!inLen || !oldSLen || !newSLen) return 0; // If any argument fails, the function fails.

// Now find occurrences of newS in input.
int oldSCount = 0;
for (i=0;i<inLen;i++) {
	for (j=0;j<oldSLen;j++) {
		if (in[i + j] != oldS[j]) break;
	}
	if (j == oldSLen) oldSCount++; // If we exited the j loop because j !< oldSLen, then OldS matches input at position i.
}

/* Find proper size of output. It equals the input length plus the product of the difference in size
 * made by each replacement and the number of replacements.
 */
size_t outLen = inLen + (newSLen - oldSLen) * oldSCount;

// Now allocate out using the proper size obtained from outLen plus one for the terminating null.
delete [] out; // Free memory to which out points, just in case, to prevent a memory leak.
out = new char[outLen + 1];

/* Do the actual work of the string replacement. Adjust the pointers themselves so there's less to keep
 * track of.
 */
while (*in != '\0') {
	for (j=0;j<oldSLen;j++) { // Just as before, check if this is a location of oldS.
		if (in[j] != oldS[j]) break;
	}
	if (j == oldSLen) { // It is a location of oldS.
		for (j=0;j<newSLen;j++) { // Copy newS to out.
			*out++ = newS[j];
		}
		in += oldSLen; // Adjust in to keep up with out.
	} else { // It's not a location of oldS.
		*out++ = *in++; // Copy in to out then increment both.
	}
}
*out = '\0'; // in should be null here (that's why the while loop was exited) so now out is, too.

out -= outLen; // Need to rewind to initial position since out is a reference to a pointer.

// Everything seems to have gone well.
return 1; // Success code
}
